<?php

declare(strict_types=1);

namespace mirzaev\vk\api\methods;

// Файлы проекта
use mirzaev\accounts\vk,
    mirzaev\vk\robots\robot,
    mirzaev\vk\api\data,
    mirzaev\vk\robots\group;

// Встроенные библиотеки
use stdClass,
    Exception;

/**
 * Режимы отправки сообщений
 */
enum mode
{
    /** Быстро - случайный идентификатор (умножение на rand()) */
    case random;

    /** Надёжно - проверка отправки (поиск сообщения через messages.getById) */
    case search;
}

/**
 * Сообщение
 *
 * @method public static function put(RobotAbstract $from, int $to, string $message, int $mode = 2) Отправить сообщение
 *
 * @see https://vk.com/dev/messages.send
 * @see https://vk.com/dev/messages.getById
 *
 * @package mirzaev\vk\api
 * @author Arsen Mirzaev Tatyano-Muradovich <arsen@mirzaev.sexy>
 *
 * @todo Доработать строгий режим отправки: проверку сообщения в беседе (не имеет ID сообщений)
 */
final class messages extends method
{
    /**
     * @var mode $mode Режим отправки сообщений
     */
    protected mode $mode = mode::random;

    /**
     * @var array[int] Сообщения для пересылки
     */
    protected array $forward;

    /**
     * @var int Сообщение для ответа
     */
    protected int $reply;

    /**
     * Конструктор
     *
     * @param robot $robot Робот
     * @param int|string|array|null $receiver Получатель
     * @param string|null $text Текст
     */
    public function __construct(
        protected robot $robot,
        int|string|array|null $receiver = null,
        protected string|null $text = null
    ) {
        if (isset($this->text, $receiver)) {
            // Быстрая отправка

            $this->send($receiver);
        }
    }

    /**
     * Записать текст
     *
     * @param string $text Текст
     */
    public function text(string $text): self
    {
        // Записать текст
        if (!isset($this->text)) {
            $this->text = $text;
        } else {
            $this->text .= $text;
        }

        return $this;
    }

    /**
     * Записать сообщения для пересылки
     *
     * @param $ids Идентификаторы сообщений
     */
    public function forward(...$ids): self
    {
        // Запись
        $this->forward = array_merge($this->forward ?? [], $ids);

        return $this;
    }

    /**
     * Записать сообщение для ответа
     *
     * @param $id Идентификатор сообщения
     */
    public function reply(int $id): self
    {
        // Запись
        $this->reply = $id;

        return $this;
    }

    /**
     * Отправить сообщение
     *
     * @see https://vk.com/dev/messages.send
     *
     * @param int|string|array|null $receiver Получатель
     *
     * @return int|array Идентификатор успешно отправленного сообщения или ответ сервера (подразумевается ошибка)
     *
     * @todo Написать обработчик ошибок возвращаемых ВКонтакте
     */
    public function send(int|string|array|null $receiver): int|array
    {
        // Идентификатор
        $random_id = time();

        if ($this->mode === mode::random) {
            // Быстрая отправка сообщения

            $random_id *= rand();
        }

        // Реиницилазиция
        $this->robot->api->reinit();

        if (is_int($receiver)) {
            // Идентификатор

            // Инициализация получателя
            $this->robot->api['peer_id'] = $receiver;
        } else if (is_array($receiver)) {
            // Идентификаторы

            // Инициализация получателя
            $this->robot->api['user_ids'] = $receiver;
        } else if (is_string($receiver)) {
            // Домен

            // Инициализация получателя
            $this->robot->api['domain'] = $receiver;
        }

        // Инициализация идентификатора сообщения (защита от повторных отправок) в настройках API
        $this->robot->api['random_id'] = $random_id;

        // Инициализация текста в настройках API
        $this->robot->api['message'] = $this->text;

        // Пересылаемые сообщения
        if (!empty($this->forwardMessages)) {

            // Инициализация пересылаемых сообщений в настройках API
            $this->robot->api['forward_messages'] = implode(',', $this->forwardMessages);
        }

        // Ответные сообщения
        if (isset($this->ReplyMessage)) {

            // Инициализация идентификатора сообщения на которое обрабатывается ответ в настройках API
            $this->robot->api['reply_to'] = $this->ReplyMessage;
        }

        // Вложения
        if (isset($this->data) && $this->__get('data') !== []) { // !empty($this->data->data) почемуто не работает

            // Инициализация вложений в настройках API
            $this->robot->api['attachment'] = implode(',', $this->__get('data'));
        }

        // Запрос
        $request = json_decode($this->robot->browser->request('POST', 'messages.send', ['form_params' => $this->robot->api->settings])->getBody()->getContents());

        // Если в ответе ошибка
        if (isset($request->error)) {
            throw new Exception('Вконтакте: ' . $request->error->error_msg, $request->error->error_code);
        }

        if ($this->mode === mode::search) {
            // Надёжная доставка сообщения

            if (!empty($request["response"])) {
                // Ответ получен

                // Реиницилазиция
                $this->robot->api->reinit();

                // Запрашиваемые сообщения
                $this->robot->api['message_ids'] = $request["response"];

                if ($this->robot->browser->request('POST', 'messages.getById', ['form_params' => $this->robot->api->settings])['response']['count'] === 0) {
                    // Сообщения не существует
                    //!!!!!!!!!!!!!!!!!!!!!!!!!!!
                    //!!!!!!!!!!!!!!!!!!!!!!!!!!
                    //!!!!!!!!!!!!!!!!!!!!!!!!!

                    // Повторная отправка
                    $this->send($receiver);
                }
            } else {
            }
        }

        return $request->response;
    }

    /**
     * Удалить сообщение
     *
     * @see https://vk.com/dev/messages.delete
     *
     * @param int|string|array|null $messages Получатель (message_ids + cmids)
     * @param int|string|null $peer_id Идентификатор беседы
     * @param bool $spam Пометить как спам?
     * @param bool $delete_for_all Удалить для всех?
     *
     * @return int|array Идентификатор успешно отправленного сообщения или ответ сервера (подразумевается ошибка)
     *
     * @todo Написать обработчик ошибок возвращаемых ВКонтакте
     */
    public function delete(string|int|array|null $messages = null, int|string|null $peer_id = null, bool $spam = false, bool $delete_for_all = false): stdClass
    {
        // Реиницилазиция настроек
        $this->robot->api->reinit();

        if (isset($peer_id)) {
            // Получен идентификатор беседы

            // Инициализация идентификатора беседы
            $this->robot->api['peer_id'] = $peer_id;

            // Инициализация идентификаторов сообщений
            $this->robot->api['cmids'] = $messages;

            // Инициализация: "удалить для всех?"
            $this->robot->api['delete_for_all'] = 1;
        } else {
            // Не получен идентификатор беседы

            // Инициализация идентификаторов сообщений
            $this->robot->api['message_ids'] = $messages;

            // Инициализация: "удалить для всех?"
            if ($delete_for_all) $this->robot->api['delete_for_all'] = 1;
        }

        // Инициализация: "сообщить о спаме?"
        if ($spam) $this->robot->api['spam'] = $spam;

        // Запрос
        $request = json_decode($this->robot->browser->request('POST', 'messages.delete', ['form_params' => $this->robot->api->settings])->getBody()->getContents());

        // Проверка на наличие ошибок в ответе от ВКонтакте
        if (isset($request->error)) throw new Exception('ВКонтакте: ' . $request->error->error_msg, $request->error->error_code);


        return $request->response;
    }

    /**
     * Записать свойство
     *
     * @param string $name Название
     * @param mixed $value Значение
     *
     * @return void
     */
    public function __set(string $name, mixed $value): void
    {
        match ($name) {
            default => throw new Exception("Свойство $name не найдено", 404)
        };
    }

    /**
     * Прочитать свойство
     *
     * @param string $name Название
     *
     * @return mixed
     */
    public function __get(string $name): mixed
    {
        return match ($name) {
            'text' => $this->text ?? throw new Exception('Текст не инициализирован'),
            'forward' => empty($this->forward) ? throw new Exception('Сообщения для пересылки не инициализированы') : $this->forward,
            'reply' => isset($this->reply) ? $this->reply : throw new Exception('Сообщение для ответа не инициализировано'),
            default => throw new Exception("Свойство $name не найдено", 404)
        };
    }

    /**
     * Получить информацию о сообщении по id
     *
     * @param string $message_ids Идентификатор сообщения
     *
     * @return object Информация о сообщении
     */

    //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    // доделать

    public function getById(int $message_ids): object
    {
        // Реиницилазиция
        $this->robot->api->reinit();

        // Идентификаторы сообщений. Максимум 100 идентификаторов
        $this->robot->api['message_ids'] = $message_ids;


        if ($this->robot instanceof group) {
            // Если это группа

            // Идентификатор группы
            isset($this->robot->id) ? $this->robot->api['group_id'] = $this->robot->id : throw new Exception('Необходимо указать идентификатор сообщества');
        }

        // Запрос
        $request = json_decode(($this->robot->browser->request('POST', 'messages.getById', ['form_params' => $this->robot->api->settings])->getBody()->getContents()));

        // Если в ответе ошибка
        if (isset($request->error)) {
            throw new Exception('Вконтакте: ' . $request->error->error_msg, $request->error->error_code);
        }

        return $request->response;
    }
}
