<?php

declare(strict_types=1);

namespace mirzaev\vk\api\methods;

use Exception;

use mirzaev\vk\robots\robot,
    mirzaev\vk\robots\group;

use mirzaev\accounts\vk as account;

/**
 * Фотографии (изображения)
 *
 * @method public static function put(RobotAbstract $from, int $to, string $message, int $mode = 2) Отправить сообщение
 *
 * @see https://vk.com/dev/photos.getUploadServer
 * @see https://vk.com/dev/messages.getById
 *
 * @package mirzaev\vk\api\methods
 * @author Arsen Mirzaev Tatyano-Muradovich <arsen@mirzaev.sexy>
 *
 * @todo Добавить обработку ошибок ($request['errors];)
 */
final class photos extends method
{
    /**
     * Конструктор
     *
     * @param robot $robot Робот
     */
    public function __construct(
        protected robot $robot
    ) {
    }

    /**
     * Сохранить
     *
     * @param robot $robot Робот
     * @param int $album_id Альбом
     * @param int|null $group_id Группа
     * @param string|null $caption Описание
     * @param float|null $latitude Географическая широта (-90, 90)
     * @param float|null $longitude Географическая долгота (-180, 180)
     *
     * @see https://vk.com/dev/photos.save
     *
     * @return array|null Ответ сервера
     */
    public static function save(robot $robot, int $album_id, array $images, string $upload_url = null, string|null $caption = null, int|null $group_id = null, float|null $latitude = null, float|null $longitude = null): ?array
    {
        if (isset($robot->account)) {
            // Если инициализирован аккаунт

            // Порядок инициализации ДОЛЖЕН соответствовать порядку в документации

            // Ключ
            $settings['access_token'] = $robot->account->key;

            // Версия API
            $settings['v'] = $robot->version;

            // Альбом
            $settings['album_id'] = $album_id;

            // Группа
            if (isset($group_id)) {
                $settings['group_id'] = $group_id;
            }

            //загрузить
            $upload = self::uploadMessage($robot, $upload_url, ...$images);

            // Сервер
            $settings['server'] = $upload['server'];

            // Изображения
            $settings['photos_list'] = $upload['photos_list'];

            // Хеш
            $settings['hash'] = $upload['hash'];

            // Широта
            if (isset($latitude)) {
                $settings['latitude'] = $latitude;
            }

            // Долгота
            if (isset($longitude)) {
                $settings['longitude'] = $longitude;
            }

            // Описание
            if (isset($caption)) {
                $settings['caption'] = $caption;
            }

            return (array) $robot->browser->api('photos.save', $settings);
        }

        throw new Exception('Аккаунт не инициализирован');
    }

    /**
     * Загрузить
     *
     * @param string $images
     *
     * @return void
     *
     * @todo Вынести этот мусор за Костей
     */
    // public static function upload(string ...$images): void
    // {
    //     if (count($images) > 5) {
    //         throw new Exception('Запрещено отправлять более 5 фотографий');
    //     }
    // }

    /**
     * загрузить
     *
     * @todo Вынести этот мусор за Костей
     */
    public static function uploadMessage(robot $robot, string $upload_url, string ...$images)
    {
        return $robot->browser->api($upload_url, ...$images);
    }

    /**
     * Получить сервер для загрузки изображений
     *
     * @see https://vk.com/dev/photos.getUploadServer
     *
     * @param robot $robot Робот
     * @param int|null $album_id Альбом
     * @param int|group|null $group_id Группа
     *
     * @return array|null Ответ сервера
     */
    public static function getUploadServer(robot $robot, int|null $album_id = null, int|group|null $group_id = null): ?array
    {
        if (isset($robot->account)) {
            // Если инициализирован аккаунт

            // Порядок инициализации ДОЛЖЕН соответствовать порядку в документации API ВКонтакте

            // Ключ
            $settings['access_token'] = $robot->account->key;

            // Версия API
            $settings['v'] = $robot->version;

            // Альбом
            match (true) {
                isset($album_id) => $settings['album_id'] = $album_id,
                default => $settings['album_id'] = self::getAlbums($robot)
            };

            // Группа
            if (isset($group_id)) {
                // При неудачном чтении воспринимать как целочисленное атомарное значение
                $settings['group_id'] = $group_id->id ?? $group_id;
            }

            return (array) $robot->browser->api('photos.getUploadServer', $settings);
        }

        throw new Exception('Аккаунт не инициализирован');
    }

    /**
     * Получить альбомы
     *
     * @see https://vk.com/dev/photos.getUploadServer
     *
     * @param robot $robot Робот
     * @param array $album_ids = null Идентификаторы альбомов
     * @param int $offset = null Смещение для выборки подмножества
     * @param int $count = null Количество для возврата
     * @param bool $need_system = null Активация возврата системных альбомов
     * @param bool $need_covers = null Активация возврата поля с обложкой альбома
     * @param bool $photo_sizes = null Активация специального формата размеров фотографий
     *
     * @return array Ответ сервера
     */
    public static function getAlbums(robot $robot, array $album_ids = null, int $offset = null, int $count = null, bool $need_system = null, bool $need_covers = null, bool $photo_sizes = null): array
    {
        // Настройки
        $settings['access_token'] = $robot->key;
        $settings['v'] = $robot->version;
        $settings['owner_id'] = $robot->id;
        if (isset($album_ids)) {
            $settings['album_ids'] = implode(',', $album_ids);
        }
        if (isset($offset)) {
            $settings['offset'] = $offset;
        }
        if (isset($count)) {
            $settings['count'] = $count;
        }
        if (isset($need_system)) {
            $settings['need_system'] = (int) $need_system;
        }
        if (isset($need_covers)) {
            $settings['need_covers'] = (int) $need_covers;
        }
        if (isset($photo_sizes)) {
            $settings['photo_sizes'] = (int) $photo_sizes;
        }

        // Запрос
        $request = $robot->browser->request('POST', 'photos.getUploadServer', $settings)->getBody()->getContents();

        return (array) $request;
    }

    /**
     * Получить адрес сервера сообщений
     *
     * @return object
     */
    public function getMessageServer(): object
    {
        // Реиницилазиция
        $this->robot->api->reinit();

        // Идентификатор группы
        $this->robot->api['group_id'] = $this->robot->id;

        // Идентификатор назначения (0 Для ботов)
        $this->robot->api['peer_id'] = 0;

        // Получить адрес сервера
        $request = json_decode($this->robot->browser->request('POST', 'photos.getMessagesUploadServer', [
            'form_params' => $this->robot->api['settings']
        ])->getBody()->getContents());

        // Если в ответе ошибка
        if (isset($request->error)) {
            throw new Exception('Вконтакте: ' . $request->error->error_msg, $request->error->error_code);
        }

        return $request->response;
    }

    /**
     * Загрузить фото
     */
    public function upload($photo, string $url): object
    {
        // Загрузить фото
        $request = json_decode($this->robot->browser->request('POST', $url, [
            'multipart' => [
                [
                    'Content-type' => 'multipart/form-data',
                    'name' => 'photo',
                    'contents' => $photo
                ]
            ]
        ])->getBody()->getContents());

        // Если в ответе ошибка
        if (isset($request->error)) {
            throw new Exception('Вконтакте: ' . $request->error->error_msg, $request->error->error_code);
        }

        return $request;
    }

    /**
     * Получить id фото для сообщения
     *
     * $robot робот
     *
     * $pathPhoto Путь к фото
     */
    public function getPhoto(string $pathPhoto): string
    {
        // Получить адрес сервера сообщений
        $url = $this->getMessageServer()->upload_url;

        // Открыть фото
        $pathPhoto = fopen($pathPhoto, 'r');

        //Загрузить фото
        $response = $this->upload($pathPhoto, $url);

        // Реинициализация
        $this->robot->api->reinit();

        // Идентификатор группы
        $this->robot->api['group_id'] = $this->robot->id;


        // Сервер
        $this->robot->api['server'] = $response->server;

        // Фото
        $this->robot->api['photo'] = $response->photo;

        // Хэш
        $this->robot->api['hash'] = $response->hash;

        // Сохранить фото
        $request = json_decode($this->robot->browser->request('POST', 'photos.saveMessagesPhoto', [
            'form_params' => $this->robot->api['settings']
        ])->getBody()->getContents());

        // Если в ответе ошибка
        if (isset($request->error)) {
            throw new Exception('Вконтакте: ' . $request->error->error_msg, $request->error->error_code);
        }
        $request;
        // Ссылка на фото
        return 'photo' . $request->response[0]->owner_id . '_' . $request->response[0]->id;
    }
}
