<?php

declare(strict_types=1);

namespace mirzaev\vk\api;

use Throwable;
use Exception;
use ArrayAccess;

use mirzaev\vk\robots\robot;

/**
 * Настройки соединения с API
 *
 * @var robot $robot Робот
 * @var array $settings Настройки
 * @var float $version Версия API
 *
 * @todo
 * 1. Создать __isset(), __unset()
 *
 * @package mirzaev\vk\api
 * @author Arsen Mirzaev Tatyano-Muradovich <arsen@mirzaev.sexy>
 */
class settings implements ArrayAccess
{
    /**
     * Версия API ВКонтакте по умолчанию
     *
     * Должна иметь тип string потому, что PHP при стандартных настройках удаляет нули у float
     */
    protected const VK_API_VERSION_DEFAULT = '5.130';

    /**
     * Конструктор
     */
    public function __construct(
        protected robot $robot,
        protected array $settings = []
    ) {
        if (empty($settings)) {
            // Настройки не получены

            // Инициализация
            $this->_init();
        }
    }

    /**
     * Инициализация (безопасная)
     *
     * @var float $version Версия API (переопределять не рекомендуется)
     */
    public function init(string $version = self::VK_API_VERSION_DEFAULT): self
    {
        // Инициализация
        static $blocked = false;

        if ($blocked) {
            // Блокировка найдена

            throw new Exception('Повторная инициализация запрещена', 500);
        }

        // Блокировка
        $blocked = true;

        // Инициализация
        try {
            $this->_init($version);
        } catch (Throwable $t) {
            throw new Exception('Не удалось инициализировать API', 500, $t->getPrevious());
        }

        return $this;
    }

    /**
     * Реинициализация
     *
     * @param bool $purge Полная очистка
     */
    public function reinit(bool $purge = false): self
    {
        // Реинициализация
        try {
            if ($purge) {
                // Полная очистка

                // Очистка
                $this->settings = [];
            } else {
                // Инициализация буфера
                $buffer = [];

                if (isset($this->settings['v'])) {
                    // Найдена версия API (подразумевается использование робота-группы)

                    // Запись в буфер
                    $buffer['v'] = $this->settings['v'];
                }

                if (isset($this->settings['access_token'])) {
                    // Найден ключ группы (подразумевается использование робота-группы)

                    // Запись в буфер
                    $buffer['access_token'] = $this->settings['access_token'];
                }

                if (isset($this->settings['group_id'])) {
                    // Найден идентификатор группы (подразумевается использование робота-группы)

                    // Запись в буфер
                    $buffer['group_id'] = $this->settings['group_id'];
                }

                if (isset($this->settings['user_id'])) {
                    // Найден идентификатор пользователя (подразумевается использование робота-пользователя)

                    // Запись в буфер
                    $buffer['user_id'] = $this->settings['user_id'];
                }

                // Перенос данных из буфера
                $this->settings = $buffer;
            }
        } catch (Throwable $t) {
            throw new Exception('Не удалось инициализировать API', 500, $t->getPrevious());
        }

        return $this;
    }

    /**
     * Инициализация
     *
     * @var float $version Версия API (переопределять не рекомендуется)
     */
    protected function _init(string $version = self::VK_API_VERSION_DEFAULT): self
    {
        // Версия API
        $this->settings['v'] = $version;

        return $this;
    }

    /**
     * Записать свойство
     *
     * @param string $name Название
     * @param mixed $value Значение
     */
    public function __set(string $name, mixed $value): void
    {
        match ($name) {
            'settings' => isset($this->settings) ? throw new Exception('Запрещено перезаписывать настройки', 500) : $this->settings = $value,
            'robot' => isset($this->robot) ? throw new Exception('Запрещено перезаписывать Робота', 500) : $this->robot = $value,
            'data', 'attachments' => $this->offsetSet('attachments', $value),
            default => $this->offsetSet($name, $value)
        };
    }

    /**
     * Прочитать свойство
     *
     * @param string $name Название
     */
    public function __get(string $name): mixed
    {
        return match ($name) {
            'settings' => $this->settings ?? throw new Exception('Настройки не инициализированы', 500),
            'robot' => $this->robot ?? throw new Exception('Робот не инициализирован', 500),
            'data', 'attachments' => $this->offsetGet('attachments'),
            default => $this->offsetGet($name)
        };
    }

    // public function __unset(string $name): void
    // {
    //     match ($name) {
    //         'settings' => throw new Exception('Запрещено удалять настройки', 500),
    //         'robot' =>  throw new Exception('Запрещено удалять робота', 500),
    //         'data', 'attachments' => $this->offsetUnset('attachments'),
    //         default => $this->offsetUnset($name)
    //     };
    // }

    /**
     * Записать по смещению
     */
    public function offsetSet(mixed $offset, mixed $value): void
    {
        if (isset($this->settings)) {
            if (strcasecmp($offset, 'settings') === 0) {
                // Полная запись

                throw new Exception('Запрещено перезаписывать настройки', 500);
            } else if (
                (strcasecmp($offset, 'data') === 0)
                || (strcasecmp($offset, 'attachment') === 0)
                || (strcasecmp($offset, 'attachments') === 0)
            ) {
                // Записать вложения

                if (is_array($value)) {
                    // Жесткая запись

                    $this->settings['attachments'] = $value;

                    return;
                }

                if (empty($this->settings['attachments']) || count($this->settings['attachments']) < 10) {
                    // Записано менее чем 10 вложений (от 0 до 9)

                    // Запись (конкатенация)
                    $this->settings['attachments'][] = $value;

                    return;
                }

                throw new Exception('Превышено ограничение на 10 вложений', 500);
            } else {
                // Запись по ключу или смещению

                $this->settings[$offset] = $value;

                return;
            }
        }

        throw new Exception('Настройки не инициализированы', 500);
    }

    /**
     * Прочитать по смещению
     */
    public function &offsetGet(mixed $offset): mixed
    {
        if (isset($this->settings)) {
            if (strcasecmp($offset, 'settings') === 0) {
                // Полное чтение

                return $this->settings;
            } else if (strcasecmp($offset, 'data') === 0 || strcasecmp($offset, 'attachments') === 0) {
                // Прочитать вложения

                return $this->settings['attachments'];
            } else if (array_key_exists($offset, $this->settings)) {
                // Прочитать по ключу или смещению

                return $this->settings[$offset];
            }

            throw new Exception(is_int($offset) ? "Смещение $offset не найдено" : "Ключ $offset не найден", 404);
        }

        throw new Exception('Настройки не инициализированы', 500);
    }

    /**
     * Проверка существования смещения
     */
    public function offsetExists(mixed $offset): bool
    {
        if (isset($this->settings)) {
            if (strcasecmp($offset, 'settings') === 0) {
                // Полная проверка

                return isset($this->settings);
            } else if (strcasecmp($offset, 'data') === 0 || strcasecmp($offset, 'attachments') === 0) {
                // Проверка вложений

                return array_key_exists('attachments', $this->settings);
            } else {
                // Проверка по ключу или смещению

                return array_key_exists($offset, $this->settings);
            }
        }

        throw new Exception('Настройки не инициализированы', 500);
    }

    /**
     * Удалить по смещению
     */
    public function offsetUnset(mixed $offset): void
    {
        if (isset($this->settings)) {
            if (strcasecmp($offset, 'settings') === 0) {
                // Полное удаление

                unset($this->settings);

                return;
            } else if (strcasecmp($offset, 'data') === 0 || strcasecmp($offset, 'attachments') === 0) {
                // Удаление вложений

                unset($this->settings['attachments']);

                return;
            } else {
                // Удаление по ключу или смещению

                unset($this->settings[$offset]);

                return;
            }
        }

        throw new Exception('Настройки не инициализированы', 500);
    }
}
