<?php

declare(strict_types=1);

namespace mirzaev\vk\robots;

use Exception;

use mirzaev\vk\core,
    mirzaev\vk\proxies\proxy,
    mirzaev\vk\captcha\captcha,
    mirzaev\vk\api\settings as api,
    mirzaev\vk\api\methods\method;

use mirzaev\accounts\vk as account;

use GuzzleHttp\Client as browser;

/**
 * Робот
 *
 * @var int $id Идентификатор
 * @var int $session Сессия
 * @var string $key Ключ
 * @var api $api API ВКонтакте
 * @var account $account Аккаунт
 * @var browser $browser Браузер
 *
 * @var int $messages_mode Режим отправки сообщений
 *
 * @method public function __construct(int $id = null) Конструктор
 * @method public function key(string $key) Инициализация ключа
 * @method public function account(account $account) Инициализация аккаунта
 * @method public function __set($name, $value) Запись свойства
 * @method public function __get($name) Чтение свойства
 * @method public function __isset($name) Проверка на инициализированность свойства
 * @method public function __call(string $method, array $params) Вызов метода
 * @method public static function __callStatic(string $method, array $params) Вызов статического метода
 * @method public function __toString() Конвертация в строку
 *
 * @package mirzaev\vk\robots
 * @author Arsen Mirzaev Tatyano-Muradovich <arsen@mirzaev.sexy>
 */
abstract class robot
{
    /**
     * @var int $session Сессия
     */
    protected readonly int $session;

    /**
     * @var account $account Аккаунт ВКонтакте
     */
    private readonly account $account;

    /**
     * @var api $api API ВКонтакте
     */
    protected readonly api $api;

    /**
     * @var browser $browser Браузер
     */
    protected readonly browser $browser;

    /**
     * Конструктор
     *
     * @param int|null $id Идентификатор робота
     * @param string|null $key Ключ API ВКонтакте
     */
    public function __construct(
        protected ?int $id = null,
        protected ?string $key = null
    ) {
        // Инициализация ядра
        $core = core::init();

        // Идентификация робота
        $this->id = $id ?? $core->robots + 1;

        // Регистрация робота в ядре
        $core->write($this->id, $this);

        // Идентификация сессии робота
        $this->session = count($core->read($this->id));

        // Инициализация настроек API
        $this->api();
    }


    /**
     * Инициализация ключа
     *
     * @param string|null $key Ключ
     *
     * @return string|self Ключ при чтении или инстанцию робота при записи
     */
    public function key(string|null $key = null): string|self
    {
        if (isset($key)) {
            // Ключ передан (подразумевается как запись)

            if (isset($this->key)) {
                // Ключ уже инициализирован

                throw new Exception('Запрещено перезаписывать ключ', 500);
            } else {
                // Ключ ещё не инициализирован

                // Запись ключа
                $this->key = $key;

                // Инициализация данных в настройках API
                $this->api()['access_token'] = $key;
            }

            return $this;
        }

        // Ключ не передан (подразумевается как чтение)

        return $this->key;
    }

    /**
     * Инициализация аккаунта
     *
     * @param account $account Аккаунт
     *
     * @return account|self Аккаунт при чтении или инстанцию робота при записи
     */
    public function account(account|null $account = null): account|self
    {
        if (isset($account)) {
            // Аккаунт передан (подразумевается как запись)

            if (isset($this->account)) {
                // Аккаунт уже инициализирован

                throw new Exception('Запрещено перезаписывать аккаунт', 500);
            } else {
                // Аккаунт ещё не инициализирован

                // Запись аккаунта
                $this->account = $account;
            }

            return $this;
        }

        // Аккаунт не передан (подразумевается как чтение)

        return $this->account;
    }

    /**
     * Инициализация API
     *
     * @return api Инстанция настроек API
     */
    public function api(): api
    {
        if (!isset($this->api)) {
            // Настройки API ещё не инициализированы

            // Запись инстанции настроек API
            $this->api = new api($this);
        }

        return $this->api;
    }

    /**
     * Записать свойство
     *
     * @param string $name Название
     * @param mixed $value Значение
     *
     * @return void
     */
    public function __set(string $name, mixed $value): void
    {
        match ($name) {
            'id' => empty($this->id) ?: $this->id = $value,
            'session' => empty($this->session) ?: $this->session = $value,
            'key' => $this->key((string) $value),
            'account' => $this->account($value),
            'api' => $this->api(),
            'browser' => empty($this->browser) ?: $this->browser = $value,
            'messages_new' => $this->messages_new = (int) $value
        };
    }

    /**
     * Прочитать свойство
     *
     * @param string $name Название
     *
     * @return mixed
     */
    public function __get(string $name): mixed
    {
        return match ($name) {
            'id' => $this->id ?? throw new Exception('Идентификатор не инициализирован', 500),
            'session' => $this->session ?? throw new Exception('Сессия не инициализирована', 500),
            'key' => $this->key(),
            'account' => $this->account(),
            'api' => $this->api(),
            'browser' => $this->browser ?? $this->browser = new browser([
                'base_uri'  => 'https://api.vk.com/method/',
                'cookies'   => true
            ]),
            'messages_new' => $this->messages_new,
            default => throw new Exception("Свойство $name не найдено", 404)
        };
    }

    /**
     * Проверить на инициализированность
     *
     * @param mixed $name Название
     */
    public function __isset(string $name): bool
    {
        return match ($name) {
            'id' => isset($this->id),
            'key' => isset($this->key),
            default => false
        };
    }

    /**
     * Вызвать метод
     *
     * Ищет класс описывающий метод API ВКонтакте,
     * создаёт и возвращает его объект
     *
     * @param string $method Метод
     * @param array $params Параметры
     *
     * @return method
     */
    public function __call(string $method, array $params): method
    {
        if (class_exists($class = '\\mirzaev\\vk\\api\\methods\\' . $method . 's')) {
            // Если найден класс реализующий запрошенный метод
            return new $class($this, ...$params);
        }

        throw new Exception("Метод $method не найден", 404);
    }

    /**
     * Вызвать статический метод
     *
     * Ищет класс описывающий метод API ВКонтакте,
     * создаёт и возвращает его объект
     *
     * @param string $method Метод
     * @param array $params Параметры
     *
     * @return method
     */
    public static function __callStatic(string $method, array $params): method
    {
        if (class_exists($class = '\\mirzaev\\vk\\api\\methods\\' . $method . 's')) {
            return $class(self, ...$params);
        }

        throw new Exception("Метод $method не найден", 404);
    }

    /**
     * Конвертировать в строку
     *
     * @return string
     */
    public function __toString(): string
    {
        return (string) $this->id;
    }
}
