<?php

declare(strict_types=1);

use PHPUnit\Framework\TestCase;

use mirzaev\vk\core;
use mirzaev\vk\robots\robot;
use mirzaev\vk\tests\settings;

use mirzaev\accounts\vk as account;

use Exception as RealException;

/**
 * @testdox API ВКонтакте
 */
final class apiTest extends TestCase
{
    use settings;

    /**
     * @var core $core Ядро фреймворка
     */
    private static core $core;

    /**
     * @var account $account Аккаунт
     */
    protected static account $account;

    /**
     * @var robot $robot Робот
     */
    private static robot $robot;

    /**
     * @testdox Инициализация аккаунта
     * @beforeClass
     */
    public function testAccountInit(): void
    {
        // Проверка входных данных
        if (empty(self::$project_id)) {
            self::markTestSkipped('Не инициализирован идентификатор проекта');
        }

        // Инициализация аккаунта
        self::$account = new account(empty(self::$id) ? rand(0, 10) : self::$id, self::$path_accounts);

        // Запись режима SSL-протокола
        self::$account->ssl = self::$ssl ?? true;

        if (empty(self::$key)) {
            // Если не указан ключ

            // Проверка входных данных
            if (empty(self::$login)) {
                self::markTestSkipped('Не инициализирован входной аккаунта');
            } else if (empty(self::$password)) {
                self::markTestSkipped('Не инициализирован пароль аккаунта');
            }

            // Деаутентификация (на всякий случай)
            self::$account->deauth();

            // Аутентификация и генерация ключа
            self::$key = self::$account->auth(self::$login, self::$password)->key(self::$project_id);
        }

        $this->assertNotNull(self::$key, 'Ошибка при инициализации ключа аккаунта');
    }

    /**
     * @testdox Деинициализация аккаунта
     * @afterClass
     */
    public static function testAccountDeinit(): void
    {
        // Инициализация аккаунта
        self::$account = new account(self::$id ?? rand(0, 10), self::$path_accounts);

        // Деаутентификация
        self::$account->deauth();
    }

    /**
     * @testdox Инициализация робота-группы
     * @before
     */
    public function testRobotGroupInit(): void
    {
        // Инициализация ядра
        self::$core = core::init();

        // Сохранение количества роботов
        $count = self::$core->robots;

        // Инициализация робота
        self::$robot = self::$core->group(self::$group_id ?? rand(0, 10));

        // Проверка
        $this->assertSame(self::$core->robots, $count + 1, 'Ошибка при инициализации робота-группы');
    }

    /**
     * @testdox Деинициализация робота-группы
     * @after
     */
    public function testRobotGroupDeinit(): void
    {
        // Очистка реестра
        self::$core->delete();

        // Проверка выброса исключения (НЕ РАБОТАЕТ, ВЕРОЯТНО БАГ)
        // $this->expectExceptionMessage('Робот с идентификатором ' . self::$robot->id . ' не найден');
        // $this->expectExceptionCode(404);

        try {
            // Чтение
            self::$core->read(self::$robot->id);
        } catch (RealException $e) {
            // Проверка
            $this->assertSame('Робот с идентификатором ' . self::$robot->id . ' не найден', $e->getMessage(), 'Не удалось удалить робота');
        }
    }

    /**
     * @testdox Инициализация
     */
    public function testInit(): void
    {
        // Инициализация
        $settings = self::$robot->key(self::$group_key)->api->init(self::$version);

        // Проверка
        $this->assertNotNull($settings['access_token']);
        $this->assertNotNull($settings['v']);
        $this->assertSame($settings['v'], self::$version);

        // Проверка
        $this->expectExceptionMessage('Повторная инициализация запрещена');

        // Инициализация (повторная)
        $settings = self::$robot->api->init(self::$version);
    }

    /**
     * @testdox Реинициализация
     */
    public function testReinit(): void
    {
        // Реинициализация
        self::$robot->key(self::$group_key)->api->reinit();
    }

    /**
     * @testdox Запись и чтение параметра
     */
    public function testWriteAndReadParameter(): void
    {
        // Запись
        self::$robot->key(self::$group_key)->api['key'] = 'value';

        // Проверка
        $this->assertSame('value', self::$robot->api['key'], 'Не удалось записать или прочитать параметр');
    }

    /**
     * @testdox Чтение неинициализированного параметра
     */
    public function testReadUndefinedParameter(): void
    {
        // Проверка выброса исключения
        $this->expectExceptionMessage('Ключ key не найден');

        // Чтение
        self::$robot->key(self::$group_key)->api['key'];
    }

    /**
     * @testdox Выбор получателя по идентификатору
     */
    public function testChooseDestinationById(): void
    {
        // Выбор получателя
        self::$robot->key(self::$group_key)->api->destination(self::$target_id);

        // Проверка
        $this->assertSame(self::$target_id, self::$robot->api['peer_id'], 'Не удалось выбрать получателя по идентификатору');
    }

    /**
     * @testdox Выбор получателей по идентификаторам
     */
    public function testChooseDestinationsByIds(): void
    {
        // Выбор получателей
        self::$robot->key(self::$group_key)->api->destination([self::$target_id, self::$target_id]);

        // Проверка
        $this->assertSame([self::$target_id, self::$target_id], self::$robot->api['user_ids'], 'Не удалось выбрать получателей по идентификаторам');
    }

    /**
     * @testdox Выбор получателя по домену
     */
    public function testChooseDestinationByDomain(): void
    {
        // Выбор получателя
        self::$robot->key(self::$group_key)->api->destination(self::$target_domain);

        // Проверка
        $this->assertSame(self::$target_domain, self::$robot->api['domain'], 'Не удалось выбрать получателя по домену');
    }

    /**
     * @testdox Запись, чтение и удаление вложений
     */
    public function testWriteAndReadAttachments(): void
    {
        // Инициализация
        self::$robot->key(self::$group_key);

        // Запись
        self::$robot->api['attachments'] = 'photo0000_0000';
        self::$robot->api['attachments'] = 'video0000_0000';
        self::$robot->api['attachments'] = 'audio0000_0000';

        // Проверка
        $this->assertSame(['photo0000_0000', 'video0000_0000', 'audio0000_0000'], self::$robot->api['attachments'], 'Не удалось записать или прочитать вложения');

        // Удаление одного вложения
        unset(self::$robot->api['attachments'][0]);

        // Реинициализация смещений (для точной проверки)
        self::$robot->api['attachments'] = array_values(self::$robot->api['attachments']);

        // Проверка
        $this->assertSame(['video0000_0000', 'audio0000_0000'], self::$robot->api['attachments'], 'Не удалось удалить или прочитать вложения');

        // Удаление всех вложений
        unset(self::$robot->api['attachments']);

        // Проверка
        $this->assertNull(self::$robot->api['attachments'], 'Не удалось удалить вложения');
    }

    /**
     * @testdox Запись вложений с превышением ограничения
     */
    public function testWriteWithOverflowAttachments()
    {
        // Проверка выброса исключения
        $this->expectExceptionMessage('Превышено ограничение на 10 вложений');

        // Инициализация
        self::$robot->key(self::$group_key);

        // Запись (11 вложений)
        self::$robot->api['attachments'] = 'photo0000_0000';
        self::$robot->api['attachments'] = 'video0000_0000';
        self::$robot->api['attachments'] = 'audio0000_0000';
        self::$robot->api['attachments'] = 'photo0000_0000';
        self::$robot->api['attachments'] = 'video0000_0000';
        self::$robot->api['attachments'] = 'audio0000_0000';
        self::$robot->api['attachments'] = 'photo0000_0000';
        self::$robot->api['attachments'] = 'video0000_0000';
        self::$robot->api['attachments'] = 'audio0000_0000';
        self::$robot->api['attachments'] = 'photo0000_0000';
        self::$robot->api['attachments'] = 'video0000_0000';
    }
}
