<?php

declare(strict_types=1);

use PHPUnit\Framework\TestCase;

use mirzaev\vk\core;
use mirzaev\vk\robots\robot;
use mirzaev\vk\tests\settings;

use mirzaev\accounts\vk as account;
use mirzaev\accounts\vk;

/**
 * @testdox [МЕТОД] Сообщения
 */
final class messagesTest extends TestCase
{
    use settings;

    /**
     * @var core $core Ядро фреймворка
     */
    private static core $core;

    /**
     * @var account $account Аккаунт
     */
    protected static account $account;

    /**
     * @var robot $robot Робот
     */
    private static robot $robot;

    /**
     * @testdox Инициализация аккаунта
     * @beforeClass
     */
    public function testAccountInit(): void
    {
        // Проверка входных данных
        if (empty(self::$project_id)) {
            self::markTestSkipped('Не инициализирован идентификатор проекта');
        }

        // Инициализация аккаунта
        self::$account = new account(empty(self::$id) ? rand(0, 10) : self::$id, self::$path_accounts);

        // Запись режима SSL-протокола
        self::$account->ssl = self::$ssl ?? true;

        if (empty(self::$key)) {
            // Если не указан ключ

            // Проверка входных данных
            if (empty(self::$login)) {
                self::markTestSkipped('Не инициализирован входной аккаунта');
            } else if (empty(self::$password)) {
                self::markTestSkipped('Не инициализирован пароль аккаунта');
            }

            // Деаутентификация (на всякий случай)
            self::$account->deauth();

            // Аутентификация и генерация ключа
            self::$key = self::$account->auth(self::$login, self::$password)->key(self::$project_id);
        }

        $this->assertNotNull(self::$key, 'Ошибка при инициализации ключа аккаунта');
    }

    /**
     * @testdox Деинициализация аккаунта
     * @afterClass
     */
    public static function testAccountDeinit(): void
    {
        // Инициализация аккаунта
        self::$account = new account(empty(self::$id) ? rand(0, 10) : self::$id, self::$path_accounts);

        // Деаутентификация
        self::$account->deauth();
    }

    /**
     * @testdox Инициализация робота
     * @before
     */
    public function testRobotGroupInit(): void
    {
        // Инициализация ядра
        self::$core = core::init();

        // Сохранение количества роботов
        $count = self::$core->robots;

        // Инициализация робота
        self::$robot = self::$core->group(empty(self::$group_id) ? rand(0, 10) : self::$group_id);

        // Проверка
        $this->assertSame(self::$core->robots, $count + 1, 'Ошибка при инициализации робота');
    }

    /**
     * @testdox Деинициализация робота
     * @after
     */
    public function testRobotGroupDeinit(): void
    {
        // Очистка реестра
        self::$core->delete();

        // Проверка
        $this->assertEmpty(self::$core->read(self::$robot->id), 'Ошибка при деинициализации робота');
    }

    /**
     * @testdox Чтение текста
     */
    public function testReadText(): void
    {
        // Проверка выброса исключения
        $this->expectExceptionMessage('Текст не инициализирован');

        // Чтение текста
        self::$robot->message()->text;
    }

    /**
     * @testdox Запись текста
     */
    public function testWriteText(): void
    {
        // Запись текста
        $message = self::$robot->message()->text('text');

        // Проверка
        $this->assertNotNull($message->text);
    }

    /**
     * @testdox Запись текста (повторная)
     */
    public function testWriteTextWhenHeIsAlreadyWrited(): void
    {
        // Запись текста
        $message = self::$robot->message()->text('hello');

        // Повторная запись текста
        $message->text(' world');

        // Проверка
        $this->assertSame('hello world', $message->text, );
    }

    /**
     * @testdox Отправка текста
     */
    public function testSendText(): void
    {
        // Отправка сообщения
        $response = self::$robot->key(self::$group_key)->message('Тестовое сообщение')->send(self::$target_id);

        // Проверка
        $this->assertIsInt($response, 'Сообщение не отправленно');
    }

    /**
     * @testdox Получение информации о сообщении
     */
    public function testGetById(): void
    {
        // Отправка сообщения
        $id = self::$robot->key(self::$group_key)->message('Теста метода getByID')->send(self::$target_id);

        // Получение информации
        $response = self::$robot->message()->getById($id);

        // Проверка
        $this->assertIsArray($response->items, 'Информация не получена');
    }


    /**
     * @testdox Чтение Вложений
     */
    public function testReadAttachments(): void
    {
        // Проверка выброса исключеия
        $this->expectExceptionMessage('Вложения не инициализированны');

        // Проверка
        self::$robot->message()->data;
    }

    /**
     * @testdox Запись вложений (методом)
     */
    public function testWriteAttachments(): void
    {
        // Запись вложений
        $message = self::$robot->message()->attachments('Вложение');

        // Проверка
        $this->assertSame(['Вложение'], $message->data);
    }

    /**
     * @testdox Запись фото
     */
    public function testWriteImage(): void
    {
        // Запись фото
        $message = self::$robot->key(self::$group_key)->message()->image(self::$path_photo);

        // Проверка
        $this->assertNotEmpty($message->data);
    }


    /**
     * @testdox Отправка фото
     */
    public function testSendPhoto(): void
    {
        // Запись фото
        $message = self::$robot->key(self::$group_key)->message('Тест отправки фото')->image(self::$path_photo);

        // Отправка сообщения
        $id = $message->send(self::$target_id);

        // Получене информации для проверки
        $message = $message->getById($id);

        // Проверка
        $this->assertNotEmpty($message->items[0]->attachments, 'Фото не отправлено');
    }

    /**
     * @testdox Чтение пересылаемых сообщений
     */
    public function testReadForward(): void
    {
        // Проверка выброса исключеия
        $this->expectExceptionMessage('Пересылаемые сообщения не инициализированны');

        // Проверка
        self::$robot->message()->forwardMessages;
    }

    /**
     * @testdox Запись пересылаемых сообщений
     */
    public function testWriteForward(): void
    {
        // Запись пересылаемых сообщений
        $message = self::$robot->message()->forward(123456);

        // Проверка
        $this->assertNotEmpty($message->forwardMessages);

    }

    /**
     * @testdox Пересылка сообщений
     */
    public function testSendForward(): void
    {
        // Отправка пересылаемого сообщения
        $id = self::$robot->key(self::$group_key)->message('Это сообщение будет переслано')->send(self::$target_id);

        // Запись пересылаемых сообщений
        $message = self::$robot->message('Тест пересылки сообщений')->forward($id);

        // Отпрравка пересылки сообщения
        $id = $message->send(self::$target_id);

        // Получене информации для проверки
        $message = self::$robot->message()->getById($id);

        // Проверка
        $this->assertNotEmpty($message->items[0]->fwd_messages, 'Сообщение не пересланно');

    }

    /**
     * @testdox Чтение ответного сообщения
     */
    public function testReadReply(): void
    {
        // Проверка выброса исключеия
        $this->expectExceptionMessage('Ответное сообщение не инициализированно');

        // Проверка
        self::$robot->message()->ReplyMessage;
    }

    /**
     * @testdox Запись ответного сообщения
     */
    public function testWriteReply(): void
    {
        // Запись ответного сообщения
        $message = self::$robot->message()->Reply(123456);

        // Проверка
        $this->assertNotNull($message->ReplyMessage);

    }

    /**
     * @testdox Ответ на сообщение
     */
    public function testSendReply(): void
    {
        // Отправка ответного сообщения
        $id = self::$robot->key(self::$group_key)->message('Будет ответ на это сообщение')->send(self::$target_id);

        // Запись ответного сообщения
        $message = self::$robot->message('Тест ответа на сообщение')->reply($id);

        // Отправка ответа на сообщение
        $id = $message->send(self::$target_id);

        // Получене информации для проверки
        $message = self::$robot->message()->getById($id);

        // Проверка
        $this->assertNotEmpty($message->items[0]->reply_message, 'Сообщение не пересланно');

    }
}
